#! /usr/bin/wish
#
# kalah.tcl : 思考ゲーム「カラー」
#
#             Copyright (C) 1998-2005,2014,2019 Makoto Hiroi
#
# 盤面 : 大域変数 board (配列）
#    12 11 10  9  8  7
#  13                  6
#     0  1  2  3  4  5
#
#     0 - 5  : my hole,   6 : my kalah
#     7 - 12 : opp hole, 13 : opp kalah
#
#  global 変数
#       board        : 盤面
#       hole_postion : 穴の座標
#       level        : 思考レベル
#       play_flag    : 6 人間の番, 13 COM の番,0 ゲーム終了
#       message      : メッセージ表示用
#       action       : 先手 0, 後手 1
#
# 思考は kalah.so をロードする
#

# 思考ルーチンロード
load ./kalah.so

# ヘルプファイルの表示
proc help {} {
    global argv0
    if {![winfo exist .t0]} {
        toplevel .t0
        wm title .t0 "Kalah Help"
        text .t0.text -yscrollcommand ".t0.scroll set"
        scrollbar .t0.scroll -command ".t0.text yview"
	pack .t0.scroll -side right -fill y
	pack .t0.text -side left
	# 起動パスを取得
	set path [file dirname $argv0]
	# ファイルの読み込み
	set f [open "$path/kalah.txt" r]
	while {![eof $f]} {
	    .t0.text insert end [read $f 1000]
	}
	close $f
	.t0.text configure -state disabled
    }
}

# ********** board を操作する関数 **********

# 駒を穴から取り出す
proc pickup_piece {pos flag} {
    global board
    set s $board($pos)
    set board($pos) 0
    if $flag {
	# 表示する
	display_hole $pos $flag
    }
    return $s
}

# 石を一つ入れるよ
proc put_piece {pos flag} {
    global board
    set board($pos) [expr $board($pos) + 1]
    if $flag {
	# 表示する
	display_hole $pos $flag
    }
}

# 石を配る turn は KALAH の位置を表す
proc distribute_piece {turn pos flag} {
    if $flag {
	display_message distribute
    }
    set num [pickup_piece $pos $flag]
    while {$num > 0} {
	incr pos 1
	# 一周したか
	if {$pos >= 14} {set pos 0}
	# 相手のカラーには入れないよ
	if {$turn ==  6 && $pos == 13} continue
	if {$turn == 13 && $pos ==  6} continue
        # 石を配る
	put_piece $pos $flag
	incr num -1
    }
    return $pos
}

# 両取りのチェック
proc check_capture_piece {turn pos flag} {
    global board
    if {($turn == 6 && $pos <= 5) || ($turn == 13 && $pos >= 7 && $pos < 13)} {
        set opp_pos [expr 12 - $pos]
        set my_piece $board($pos)
        set opp_piece $board($opp_pos)
	if {($my_piece == 1) && ($opp_piece > 0)} {
	    if $flag {
		display_message capture
	    }
	    set num [expr $my_piece + $opp_piece]
	    pickup_piece $pos $flag
	    pickup_piece $opp_pos $flag
	    set board($turn) [expr $board($turn) + $num]
	    if $flag {
		display_hole $turn $flag
	    }
	    return $num
	}
    }
    return 0
}

# 穴に残っている駒を数える
proc count_piece {turn} {
    global board
    set count 0
    for {set i [expr $turn - 6]} {$i < $turn} {incr i} {
	incr count $board($i)
    }
    return $count
}

# 穴に残っている駒をカラーに入れる
proc hole_to_kalah {turn flag} {
    global board
    if $flag {
	display_message finish
    }
    for {set i [expr $turn - 6]} {$i < $turn} {incr i} {
	if $board($i) {
	    incr board($turn) $board($i)
	    set board($i) 0
	    if $flag {
		display_hole $i 1
		display_hole $turn 1
	    }
	}
    }
}

# 終了チェック
proc check_finish {flag} {
    global board
    if {[count_piece 6] == 0} {
	hole_to_kalah 13 $flag
    } elseif {[count_piece 13 ] == 0} {
	hole_to_kalah 6 $flag
    } elseif {($board(6) <= 36) && ($board(13) <= 36)} {
	return 0
    }
    return 1
}

# ********** 表示用サブルーチン **********

# 穴の状態を表示（サブ）
proc display_hole_sub {n i x c} {
    global piece_id
    set y 4
    while {$n > 0} {
	.c$i itemconfigure $piece_id($i,$x,$y) -fill $c
        incr y -1
	incr n -1
    }
    while {$y >= 0} {
	.c$i itemconfigure $piece_id($i,$x,$y) -fill gray
	incr y -1
    }
}

# 穴の状態を表示する
proc display_hole {i c} {
    global board piece_id
    # 穴の状態を表示
    set n0 [expr $board($i) / 10]
    set n2 [expr $board($i) % 10]
    if {$n2 > 5} {
	set n1 [expr $n2 - 5]
	set n2 5
    } else {
	set n1 0
    }
    display_hole_sub $n0 $i 0 red
    display_hole_sub $n1 $i 1 blue
    display_hole_sub $n2 $i 2 blue
    update
    if {$c > 0} {
	# 500 msec 待つ
	after 500
    }
}

# メッセージの表示
proc display_message {m} {
    global message mes_table
    set message $mes_table($m)
    update
    after 500
}


# ********** オペレーション **********

proc game_over {} {
    global board play_flag
    if {$board(6) == $board(13)} {
	set m "引き分けです"
    } elseif {$board(6) > $board(13)} {
	set m "あなたの勝ちです"
    } else {
	set m "私の勝ちです"
    }
    tk_messageBox -type ok -message $m
    set play_flag 0
    display_message start
}

proc move_piece_com {} {
    global level play_flag board
    display_message opp_turn
    # kalah.dll の think を呼び出す
    . configure -cursor watch
    update
    set result [think $level 13 board]
    . configure -cursor ""
    # 指し手はリストに格納されているよ
    foreach m $result {
	set pos [distribute_piece 13 $m 1]
	check_capture_piece 13 $pos 1
        if {$pos == 13} {
	    display_message kalah
	    display_message opp_turn
	}
	if [check_finish 1] {
	    game_over
	    return
	}
    }
    display_message my_turn
    set play_flag 6
}

proc move_piece_human {i} {
    global board play_flag
    # 人間側の指し手
    if {$board($i) > 0 && $play_flag == 6} {
        # 二重入力を防ぐ
        set play_flag 13
	set pos [distribute_piece 6 $i 1]
	check_capture_piece 6 $pos 1
	if [check_finish 1] {
	    # 終了
	    game_over
	    return
	}
	if {$pos == 6} {
	    # カラーに入った
	    display_message kalah
            # 入力受付
	    set play_flag 6
	    return
	}
	# コンピュータの指し手
	move_piece_com
    }
}

# 盤面の初期化
proc init_board {} {
    global board
    for {set i 0} {$i < 14} {incr i} {
	if {$i == 6 || $i == 13} {
	    set board($i) 0
	} else {
	    set board($i) 6
	}
	display_hole $i 0
    }
}

# ゲームの開始
proc new_game {} {
    global play_flag action
    if {$play_flag == 6} {
	set ans [tk_messageBox -type yesno -icon question \
                   -message "今のゲームを中断しますか？"]
        if {$ans == "no"} return
	set play_flag 0
    }
    if {$play_flag == 0 } {
	# ゲーム開始
	init_board
	if $action {
	    set play_flag 13
	    move_piece_com
	} else {
	    set play_flag 6
	    display_message my_turn
	}
    }
}

# ********** 画面の初期化 **********

# 穴の位置を設定 大きさは 48 * 80
set hole_postion(0)  {80  168}
set hole_postion(1)  {144 168}
set hole_postion(2)  {208 168}
set hole_postion(3)  {272 168}
set hole_postion(4)  {336 168}
set hole_postion(5)  {400 168}
set hole_postion(6)  {464 92}
set hole_postion(7)  {400 16}
set hole_postion(8)  {336 16}
set hole_postion(9)  {272 16}
set hole_postion(10) {208 16}
set hole_postion(11) {144 16}
set hole_postion(12) {80  16}
set hole_postion(13) {16  92}

# 穴を作る
proc make_hole {} {
    global hole_postion piece_id
    for {set i 0} {$i < 14} {incr i} {
	set x0 [lindex $hole_postion($i) 0]
	set y0 [lindex $hole_postion($i) 1]
	canvas .c$i -width 48 -height 80 -bg gray
	place .c$i -x $x0 -y $y0
	# 駒表示用の円を１５個用意する
	for {set x 0} {$x < 3} {incr x} {
	    for {set y 0} {$y < 5} {incr y} {
		set x1 [expr $x * 16 + 2]
		set y1 [expr $y * 16 + 2]
		set x2 [expr $x1 + 14]
		set y2 [expr $y1 + 14]
		set piece_id($i,$x,$y) \
		    [.c$i create oval $x1 $y1 $x2 $y2 -outline gray]
	    }
	}
    }
}

# ラベルを作成する
proc make_label {} {
    global hole_postion board message
    for {set i 0} {$i < 14} {incr i} {
	set x0 [lindex $hole_postion($i) 0]
	set y0 [lindex $hole_postion($i) 1]
	label .l$i -textvariable board($i) -bg darkgreen -fg white
	place .l$i -x [expr $x0 + 14] -y [expr $y0 + 84]
        # カラーのラベル
	if {$i == 6 || $i == 13} {
	    label .lk$i -text "Kalah" -bg darkgreen -fg red
	    place .lk$i -x $x0 -y [expr $y0 - 28]
	}
    }
    # メッセージ表示用ラベル
    label .lm -textvariable message -bg darkgreen -fg yellow
    place .lm -x 100 -y 128
}

# グローバル変数の初期化
proc init_global {} {
    global level action mes_table play_flag message
    # 乱数の初期化
    expr srand( [clock seconds] )
    set level 1
    set action 0
    set play_flag 0
    set mes_table(start)      "メニュー Games の Start でゲーム開始！"
    set mes_table(kalah)      "カラーに入った！ もう一回どうぞ"
    set mes_table(my_turn)    "あなたの手番です"
    set mes_table(opp_turn)   "わたしの手番です"
    set mes_table(capture)    "両取りです"
    set mes_table(finish)     "残っている駒をカラーに入れます"
    set mes_table(distribute) "駒を配ります"
    set message $mes_table(start)
}

# ********** メニューの設定 **********
wm geometry . 530x310
menu .m -type menubar
. configure -menu .m -bg darkgreen
.m add cascade -label "Games" -under 0 -menu .m.m1
.m add cascade -label "Level" -under 0 -menu .m.m2
.m add command -label "Help"  -under 0 -command "help"
menu .m.m1 -tearoff no
.m.m1 add command -label "Start" -under 0 -command "new_game"
.m.m1 add separator
.m.m1 add radiobutton -label "先手" -variable action -value 0
.m.m1 add radiobutton -label "後手" -variable action -value 1
.m.m1 add separator
.m.m1 add command -label "Exit" -under 0 -command "exit"
menu .m.m2 -tearoff no
.m.m2 add radiobutton -label "Level 1" -variable level -value 1
.m.m2 add radiobutton -label "Level 2" -variable level -value 2
.m.m2 add radiobutton -label "Level 3" -variable level -value 3
.m.m2 add radiobutton -label "Level 4" -variable level -value 4
.m.m2 add radiobutton -label "Level 5" -variable level -value 5


# ********** 画面の生成 **********
option add *font "{Noto Sans Mono CJK JP} 12"
init_global
make_hole
make_label

# ********** バインド **********
for {set i 0} {$i < 6} {incr i} {
    bind .c$i <Button-1> "move_piece_human $i"
}


wm title . "Kalah"
wm resizable . 0 0
focus -force .

# ********** end of file **********
